
// Algorithm for efficiently running video game console emulator in reverse
//
// See nesdev thread: http://nesdev.parodius.com/bbs/viewtopic.php?t=331
//
// Copyright (C) 2005 Shay Green. MIT license.
//
// Shay Green <hotpop.com@blargg> (swap to e-mail)

#include "reverse_emulation.h"

typedef long timestamp_t;              // absolute frame number
static timestamp_t current_frame = -1; // currently displayed frame
static timestamp_t emulator_frame = 0; // frame emulator would generate if run
static int buf_size = 1;

void re_init( int n )
{
	buf_size = n;
	emulator_frame = 0;
	current_frame = -1;
}

static void restore_state( int n )
{
	emulator_frame = n * buf_size;
	emu_restore_state( n );
}

static void emulate_frame( int i )
{
	emulator_frame++;
	emu_run_frame( i );
}

static int buffer_scrambled;   // number of frames remaining before buffer isn't scrambled
static int reverse_mode;       // reverse mode is in effect
static int reverse_unmirrored; // number of frames until buffer order is mirrored
static int reverse_pivot;      // pivot point in buffer

static int reversed_index()
{
	int index = current_frame % buf_size;
	if ( !reverse_unmirrored )
		index = buf_size - 1 - index;
	if ( index <= reverse_pivot )
		index = reverse_pivot - index;
	return index;
}

// Run emulation backwards one frame and return buffer index to be displayed
int re_prev_frame()
{
	if ( !reverse_mode )
	{
		//printf( "Reverse mode\n" );
		
		reverse_pivot = current_frame % buf_size;
		
		if ( buffer_scrambled )
		{
			//printf( "Refill buffer\n" );
			
			// Last reverse was less than buf_size frames ago, so buffer
			// might be scrambled. Need to fill entire buffer. Won't occur
			// much so not a performance problem.
			
			// fill beginning (this can be optimized out, but not easily)
			restore_state( current_frame / buf_size );
			
			{
				int i;
				for ( i = 0; i <= reverse_pivot; i++ )
					emulate_frame( i );
			}
			
			// fill end
			restore_state( current_frame / buf_size - 1 );
			
			{
				int i;
				for ( i = 0; i < buf_size; i++ )
				{
					if ( i <= reverse_pivot )
					{
						emu_run_frame_invisibly();
						emulator_frame++;
					}
					else
					{
						emulate_frame( i );
					}
				}
			}
		}
		
		// Buffer won't have a clean history until run forward one time through
		buffer_scrambled = buf_size;
		
		restore_state( current_frame / buf_size - 1 );
		
		reverse_unmirrored = 2; // unmirrored for first two passes, then alternating
		reverse_pivot = -reverse_pivot; // don't pivot yet
		reverse_mode = 1;
	}
	
	emulate_frame( reversed_index() );
	
	current_frame--;
	if ( current_frame % buf_size == buf_size - 1 )
	{
		if ( reverse_pivot < 0 )
			reverse_pivot = -reverse_pivot;
		
		if ( --reverse_unmirrored < 0 )
			reverse_unmirrored = 1;
		
		restore_state( current_frame / buf_size - 1 );
	}
	
	return reversed_index();
}

// Run emulation forward one frame and return buffer index to be displayed
int re_next_frame()
{
	if ( reverse_mode )
	{
		//printf( "Forward mode\n" );
		
		// Emulate to point where playback will be resumed. This could
		// cause a delay on processor-hungry emulators. There's not much
		// that can be done to eliminate this.
		restore_state( current_frame / buf_size );
		{
			int n = current_frame % buf_size + 1;
			while ( n-- )
			{
				emu_run_frame_invisibly();
				emulator_frame++;
			}
		}
		
		reverse_mode = 0;
	}
	
	{
		int i = emulator_frame % buf_size;
		if ( i == 0 )
			emu_save_state( emulator_frame / buf_size );
		
		current_frame = emulator_frame;
		emulate_frame( i );
		
		if ( buffer_scrambled )
			buffer_scrambled--;
		
		return i;
	}
}

